# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
from pyomo.environ import (
    ConcreteModel,
    Var,
    Objective,
    Constraint,
    maximize,
    Expression,
    log10,
)
from pyomo.opt import SolverFactory, TerminationCondition

from pyomo.solvers.plugins.solvers.GAMS import GAMSShell, GAMSDirect, gdxcc_available
import pyomo.common.unittest as unittest
from pyomo.common.tempfiles import TempfileManager
from pyomo.common.tee import capture_output
import os, shutil
from tempfile import mkdtemp

opt_py = SolverFactory('gams', solver_io='python')
gamspy_available = opt_py.available(exception_flag=False)
if gamspy_available:
    try:
        from gams import GamsExceptionExecution
    except:
        from gams.workspace import GamsExceptionExecution

opt_gms = SolverFactory('gams', solver_io='gms')
gamsgms_available = opt_gms.available(exception_flag=False)


@unittest.pytest.mark.solver("gams")
class GAMSTests(unittest.TestCase):
    @unittest.skipIf(
        not gamspy_available, "The 'gams' python bindings are not available"
    )
    def test_check_expr_eval_py(self):
        with SolverFactory("gams", solver_io="python") as opt:
            m = ConcreteModel()
            m.x = Var()
            m.e = Expression(expr=log10(m.x) + 5)
            m.c = Constraint(expr=m.x >= 10)
            m.o = Objective(expr=m.e)

            self.assertRaises(GamsExceptionExecution, opt.solve, m)

    @unittest.skipIf(not gamsgms_available, "The 'gams' executable is not available")
    def test_check_expr_eval_gms(self):
        with SolverFactory("gams", solver_io="gms") as opt:
            m = ConcreteModel()
            m.x = Var()
            m.e = Expression(expr=log10(m.x) + 5)
            m.c = Constraint(expr=m.x >= 10)
            m.o = Objective(expr=m.e)

            self.assertRaises(ValueError, opt.solve, m)

    @unittest.skipIf(
        not gamspy_available, "The 'gams' python bindings are not available"
    )
    def test_file_removal_py(self):
        with SolverFactory("gams", solver_io="python") as opt:
            m = ConcreteModel()
            m.x = Var()
            m.c = Constraint(expr=m.x >= 10)
            m.o = Objective(expr=m.x)

            tmpdir = mkdtemp()

            results = opt.solve(m, tmpdir=tmpdir)

            self.assertTrue(os.path.exists(tmpdir))
            self.assertFalse(os.path.exists(os.path.join(tmpdir, '_gams_py_gjo0.gms')))
            self.assertFalse(os.path.exists(os.path.join(tmpdir, '_gams_py_gjo0.lst')))
            self.assertFalse(os.path.exists(os.path.join(tmpdir, '_gams_py_gdb0.gdx')))

            os.rmdir(tmpdir)

            results = opt.solve(m, tmpdir=tmpdir)

            self.assertFalse(os.path.exists(tmpdir))

    @unittest.skipIf(not gamsgms_available, "The 'gams' executable is not available")
    def test_file_removal_gms(self):
        with SolverFactory("gams", solver_io="gms") as opt:
            m = ConcreteModel()
            m.x = Var()
            m.c = Constraint(expr=m.x >= 10)
            m.o = Objective(expr=m.x)

            tmpdir = mkdtemp()

            results = opt.solve(m, tmpdir=tmpdir)

            self.assertTrue(os.path.exists(tmpdir))
            self.assertFalse(os.path.exists(os.path.join(tmpdir, 'model.gms')))
            self.assertFalse(os.path.exists(os.path.join(tmpdir, 'output.lst')))
            self.assertFalse(os.path.exists(os.path.join(tmpdir, 'GAMS_MODEL_p.gdx')))
            self.assertFalse(os.path.exists(os.path.join(tmpdir, 'GAMS_MODEL_s.gdx')))

            os.rmdir(tmpdir)

            results = opt.solve(m, tmpdir=tmpdir)

            self.assertFalse(os.path.exists(tmpdir))

    @unittest.skipIf(
        not gamspy_available, "The 'gams' python bindings are not available"
    )
    def test_keepfiles_py(self):
        with SolverFactory("gams", solver_io="python") as opt:
            m = ConcreteModel()
            m.x = Var()
            m.c = Constraint(expr=m.x >= 10)
            m.o = Objective(expr=m.x)

            tmpdir = mkdtemp()

            results = opt.solve(m, tmpdir=tmpdir, keepfiles=True)

            self.assertTrue(os.path.exists(tmpdir))
            self.assertTrue(os.path.exists(os.path.join(tmpdir, '_gams_py_gjo0.gms')))
            self.assertTrue(os.path.exists(os.path.join(tmpdir, '_gams_py_gjo0.lst')))
            self.assertTrue(os.path.exists(os.path.join(tmpdir, '_gams_py_gdb0.gdx')))
            self.assertTrue(os.path.exists(os.path.join(tmpdir, '_gams_py_gjo0.pf')))

            shutil.rmtree(tmpdir)

    @unittest.skipIf(not gamsgms_available, "The 'gams' executable is not available")
    def test_keepfiles_gms(self):
        with SolverFactory("gams", solver_io="gms") as opt:
            m = ConcreteModel()
            m.x = Var()
            m.c = Constraint(expr=m.x >= 10)
            m.o = Objective(expr=m.x)

            tmpdir = mkdtemp()

            results = opt.solve(m, tmpdir=tmpdir, keepfiles=True)

            self.assertTrue(os.path.exists(tmpdir))
            self.assertTrue(os.path.exists(os.path.join(tmpdir, 'model.gms')))
            self.assertTrue(os.path.exists(os.path.join(tmpdir, 'output.lst')))
            if gdxcc_available:
                self.assertTrue(
                    os.path.exists(os.path.join(tmpdir, 'GAMS_MODEL_p.gdx'))
                )
                self.assertTrue(os.path.exists(os.path.join(tmpdir, 'results_s.gdx')))
            else:
                self.assertTrue(os.path.exists(os.path.join(tmpdir, 'results.dat')))
                self.assertTrue(os.path.exists(os.path.join(tmpdir, 'resultsstat.dat')))

            shutil.rmtree(tmpdir)

    @unittest.skipIf(
        not gamspy_available, "The 'gams' python bindings are not available"
    )
    def test_fixed_var_sign_py(self):
        with SolverFactory("gams", solver_io="python") as opt:
            m = ConcreteModel()
            m.x = Var()
            m.y = Var()
            m.z = Var()
            m.z.fix(-3)
            m.c1 = Constraint(expr=m.x + m.y - m.z == 0)
            m.c2 = Constraint(expr=m.z + m.y - m.z >= -10000)
            m.c3 = Constraint(expr=-3 * m.z + m.y - m.z >= -10000)
            m.c4 = Constraint(expr=-m.z + m.y - m.z >= -10000)
            m.c5 = Constraint(expr=m.x <= 100)
            m.o = Objective(expr=m.x, sense=maximize)

            results = opt.solve(m)

            self.assertEqual(
                results.solver.termination_condition, TerminationCondition.optimal
            )

    @unittest.skipIf(not gamsgms_available, "The 'gams' executable is not available")
    def test_fixed_var_sign_gms(self):
        with SolverFactory("gams", solver_io="gms") as opt:
            m = ConcreteModel()
            m.x = Var()
            m.y = Var()
            m.z = Var()
            m.z.fix(-3)
            m.c1 = Constraint(expr=m.x + m.y - m.z == 0)
            m.c2 = Constraint(expr=m.z + m.y - m.z >= -10000)
            m.c3 = Constraint(expr=-3 * m.z + m.y - m.z >= -10000)
            m.c4 = Constraint(expr=-m.z + m.y - m.z >= -10000)
            m.c5 = Constraint(expr=m.x <= 100)
            m.o = Objective(expr=m.x, sense=maximize)

            results = opt.solve(m)

            self.assertEqual(
                results.solver.termination_condition, TerminationCondition.optimal
            )

    @unittest.skipIf(
        not gamspy_available, "The 'gams' python bindings are not available"
    )
    def test_long_var_py(self):
        with SolverFactory("gams", solver_io="python") as opt:
            m = ConcreteModel()
            x = m.a23456789012345678901234567890123456789012345678901234567890123 = (
                Var()
            )
            y = m.b234567890123456789012345678901234567890123456789012345678901234 = (
                Var()
            )
            z = (
                m.c23456789012345678901234567890123456789012345678901234567890123456789012345678901234567890
            ) = Var()
            w = m.d01234567890 = Var()
            m.c1 = Constraint(expr=x + y + z + w == 0)
            m.c2 = Constraint(expr=x >= 10)
            m.o = Objective(expr=x)

            results = opt.solve(m)

            self.assertEqual(
                results.solver.termination_condition, TerminationCondition.optimal
            )

    @unittest.skipIf(not gamsgms_available, "The 'gams' executable is not available")
    def test_long_var_gms(self):
        with SolverFactory("gams", solver_io="gms") as opt:
            m = ConcreteModel()
            x = m.a23456789012345678901234567890123456789012345678901234567890123 = (
                Var()
            )
            y = m.b234567890123456789012345678901234567890123456789012345678901234 = (
                Var()
            )
            z = (
                m.c23456789012345678901234567890123456789012345678901234567890123456789012345678901234567890
            ) = Var()
            w = m.d01234567890 = Var()
            m.c1 = Constraint(expr=x + y + z + w == 0)
            m.c2 = Constraint(expr=x >= 10)
            m.o = Objective(expr=x)

            results = opt.solve(m)

            self.assertEqual(
                results.solver.termination_condition, TerminationCondition.optimal
            )

    def test_subsolver_notation(self):
        opt1 = SolverFactory("gams:ipopt", solver_io="gms")
        self.assertTrue(isinstance(opt1, GAMSShell))
        self.assertEqual(opt1.options["solver"], "ipopt")

        opt2 = SolverFactory("gams:baron", solver_io="python")
        self.assertTrue(isinstance(opt2, GAMSDirect))
        self.assertEqual(opt2.options["solver"], "baron")

        opt3 = SolverFactory("py:gams")
        self.assertTrue(isinstance(opt3, GAMSDirect))
        opt3.options["keepfiles"] = True
        self.assertEqual(opt3.options["keepfiles"], True)

        opt4 = SolverFactory("py:gams:cbc")
        self.assertTrue(isinstance(opt4, GAMSDirect))
        self.assertEqual(opt4.options["solver"], "cbc")

    @unittest.skipIf(
        not gamspy_available, "The 'gams' python bindings are not available"
    )
    def test_options_py(self):
        with SolverFactory("gams", solver_io="python") as opt:
            m = ConcreteModel()
            m.x = Var()
            m.c = Constraint(expr=m.x >= 10)
            m.o = Objective(expr=m.x)

            opt.options["load_solutions"] = False  # set option
            opt.solve(m)  # use option
            self.assertEqual(m.x.value, None)
            opt.solve(m, load_solutions=True)  # overwrite option
            self.assertEqual(m.x.value, 10)

    @unittest.skipIf(not gamsgms_available, "The 'gams' executable is not available")
    def test_options_gms(self):
        with SolverFactory("gams", solver_io="gms") as opt:
            m = ConcreteModel()
            m.x = Var()
            m.c = Constraint(expr=m.x >= 10)
            m.o = Objective(expr=m.x)

            opt.options["load_solutions"] = False  # set option
            opt.solve(m)  # use option
            self.assertEqual(m.x.value, None)
            opt.solve(m, load_solutions=True)  # overwrite option
            self.assertEqual(m.x.value, 10)

    @unittest.skipIf(
        not gamspy_available, "The 'gams' python bindings are not available"
    )
    def test_version_py(self):
        with SolverFactory("gams", solver_io="python") as opt:
            self.assertIsNotNone(opt.version())

    @unittest.skipIf(not gamsgms_available, "The 'gams' executable is not available")
    def test_version_gms(self):
        with SolverFactory("gams", solver_io="gms") as opt:
            self.assertIsNotNone(opt.version())

    @unittest.skipIf(not gamsgms_available, "The 'gams' executable is not available")
    def test_dat_parser(self):
        # This tests issue 2571
        m = pyo.ConcreteModel()
        m.S = pyo.Set(initialize=list(range(5)))
        m.a_long_var_name = pyo.Var(m.S, bounds=(0, 1), initialize=1)
        m.obj = pyo.Objective(
            expr=2000 * pyo.summation(m.a_long_var_name), sense=pyo.maximize
        )
        solver = pyo.SolverFactory("gams:conopt")
        res = solver.solve(
            m,
            symbolic_solver_labels=True,
            load_solutions=False,
            io_options={'put_results_format': 'dat'},
        )
        self.assertEqual(res.solution[0].Objective['obj']['Value'], 10000)
        for i in range(5):
            self.assertEqual(
                res.solution[0].Variable[f'a_long_var_name_{i}_']['Value'], 1
            )


class GAMSLogfileTestBase(unittest.TestCase):
    def setUp(self):
        """Set up model and temporary directory."""
        m = ConcreteModel()
        m.x = Var()
        m.c = Constraint(expr=m.x >= 10)
        m.o = Objective(expr=m.x)
        self.m = m
        self.tmpdir = mkdtemp()
        self.logfile = os.path.join(self.tmpdir, 'logfile.log')
        self.characteristic_output_string = "Starting compilation"

    def tearDown(self):
        """Clean up temporary directory after tests are over."""
        shutil.rmtree(self.tmpdir)

    def _check_logfile(self, exists=True, logfile=None):
        """Check for logfiles existence and contents.

        exists=True:
            Whether to check if the logfile exists or doesn't exist.
        expected=None:
            Optionally check that the logfiles contents is equal to this value.

        """
        if not logfile:
            logfile = self.logfile
        if not exists:
            self.assertFalse(os.path.exists(logfile))
            return

        self.assertTrue(os.path.exists(logfile))
        with open(logfile) as f:
            logfile_contents = f.read()
        self.assertIn(self.characteristic_output_string, logfile_contents)

    def _check_stdout(self, output_string, exists=True):
        if exists:
            # Starting Compilation is outputted by the solver itself which in this
            # case should be printed to stdout and captured
            self.assertIn(self.characteristic_output_string, output_string)
        else:
            # Note that we do not check that the output string is completely
            # empty as certain platforms (like linux) output other lines to
            # stdout where as windows for example does not. We instead just
            # check that the characteristic string is missing.
            self.assertNotIn(self.characteristic_output_string, output_string)


@unittest.skipIf(not gamsgms_available, "The 'gams' executable is not available")
@unittest.pytest.mark.solver("gams")
class GAMSLogfileGmsTests(GAMSLogfileTestBase):
    """Test class for testing permultations of tee and logfile options.

    The tests build a simple model and solve it using the different options
    using the gams command directly.

    """

    def test_no_tee(self):
        with SolverFactory("gams", solver_io="gms") as opt:
            with capture_output() as output:
                opt.solve(self.m, tee=False)
        self._check_stdout(output.getvalue(), exists=False)
        self._check_logfile(exists=False)

    def test_tee(self):
        with SolverFactory("gams", solver_io="gms") as opt:
            with capture_output() as output:
                opt.solve(self.m, tee=True)
        self._check_stdout(output.getvalue(), exists=True)
        self._check_logfile(exists=False)

    def test_logfile(self):
        with SolverFactory("gams", solver_io="gms") as opt:
            with capture_output() as output:
                opt.solve(self.m, logfile=self.logfile)
        self._check_stdout(output.getvalue(), exists=False)
        self._check_logfile(exists=True)

    def test_logfile_with_spaces(self):
        # In response to Issue 3579
        logfile_with_spaces = os.path.join(self.tmpdir, "My File.log")
        with SolverFactory("gams", solver_io="gms") as opt:
            with capture_output() as output:
                opt.solve(self.m, logfile=logfile_with_spaces)
        self._check_stdout(output.getvalue(), exists=False)
        self._check_logfile(exists=True, logfile=logfile_with_spaces)

    def test_logfile_relative(self):
        cwd = os.getcwd()
        with TempfileManager:
            tmpdir = TempfileManager.create_tempdir()
            os.chdir(tmpdir)
            try:
                self.logfile = 'test-gams.log'
                with SolverFactory("gams", solver_io="gms") as opt:
                    with capture_output() as output:
                        opt.solve(self.m, logfile=self.logfile)
                self._check_stdout(output.getvalue(), exists=False)
                self._check_logfile(exists=True)
                self.assertTrue(os.path.exists(os.path.join(tmpdir, self.logfile)))
            finally:
                os.chdir(cwd)

    def test_tee_and_logfile(self):
        with SolverFactory("gams", solver_io="gms") as opt:
            with capture_output() as output:
                opt.solve(self.m, logfile=self.logfile, tee=True)
        self._check_stdout(output.getvalue(), exists=True)
        self._check_logfile(exists=True)


@unittest.skipIf(not gamspy_available, "The 'gams' python bindings are not available")
@unittest.pytest.mark.solver("gams")
class GAMSLogfilePyTests(GAMSLogfileTestBase):
    """Test class for testing permultations of tee and logfile options.

    The tests build a simple model and solve it using the different options
    using the python gams bindings.

    """

    def test_no_tee(self):
        with SolverFactory("gams", solver_io="python") as opt:
            with capture_output() as output:
                opt.solve(self.m, tee=False)
        self._check_stdout(output.getvalue(), exists=False)
        self._check_logfile(exists=False)

    def test_tee(self):
        with SolverFactory("gams", solver_io="python") as opt:
            with capture_output() as output:
                opt.solve(self.m, tee=True)
        self._check_stdout(output.getvalue(), exists=True)
        self._check_logfile(exists=False)

    def test_logfile(self):
        with SolverFactory("gams", solver_io="python") as opt:
            with capture_output() as output:
                opt.solve(self.m, logfile=self.logfile)
        self._check_stdout(output.getvalue(), exists=False)
        self._check_logfile(exists=True)

    def test_logfile_with_spaces(self):
        # In response to Issue 3579
        logfile_with_spaces = os.path.join(self.tmpdir, "My File.log")
        with SolverFactory("gams", solver_io="gms") as opt:
            with capture_output() as output:
                opt.solve(self.m, logfile=logfile_with_spaces)
        self._check_stdout(output.getvalue(), exists=False)
        self._check_logfile(exists=True, logfile=logfile_with_spaces)

    def test_logfile_relative(self):
        cwd = os.getcwd()
        with TempfileManager:
            tmpdir = TempfileManager.create_tempdir()
            os.chdir(tmpdir)
            try:
                self.logfile = 'test-gams.log'
                with SolverFactory("gams", solver_io="python") as opt:
                    with capture_output() as output:
                        opt.solve(self.m, logfile=self.logfile)
                self._check_stdout(output.getvalue(), exists=False)
                self._check_logfile(exists=True)
                self.assertTrue(os.path.exists(os.path.join(tmpdir, self.logfile)))
            finally:
                os.chdir(cwd)

    def test_tee_and_logfile(self):
        with SolverFactory("gams", solver_io="python") as opt:
            with capture_output() as output:
                opt.solve(self.m, logfile=self.logfile, tee=True)
        self._check_stdout(output.getvalue(), exists=True)
        self._check_logfile(exists=True)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common import unittest
import pyomo.environ as pyo
from pyomo.contrib.appsi.utils import collect_vars_and_named_exprs
from pyomo.contrib.appsi.cmodel import cmodel, cmodel_available
from typing import Callable
from pyomo.common.gsl import find_GSL


class TestCollectVarsAndNamedExpressions(unittest.TestCase):
    def basics_helper(self, collector: Callable, *args):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.z = pyo.Var()
        m.E = pyo.Expression(expr=2 * m.z + 1)
        m.y.fix(3)
        e = m.x * m.y + m.x * m.E
        named_exprs, var_list, fixed_vars, external_funcs = collector(e, *args)
        self.assertEqual([m.E], named_exprs)
        self.assertEqual([m.x, m.y, m.z], var_list)
        self.assertEqual([m.y], fixed_vars)
        self.assertEqual([], external_funcs)

    def test_basics(self):
        self.basics_helper(collect_vars_and_named_exprs)

    @unittest.skipUnless(cmodel_available, 'appsi extensions are not available')
    def test_basics_cmodel(self):
        self.basics_helper(cmodel.prep_for_repn, cmodel.PyomoExprTypes())

    def external_func_helper(self, collector: Callable, *args):
        DLL = find_GSL()
        if not DLL:
            self.skipTest('Could not find amplgsl.dll library')

        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.z = pyo.Var()
        m.hypot = pyo.ExternalFunction(library=DLL, function='gsl_hypot')
        func = m.hypot(m.x, m.x * m.y)
        m.E = pyo.Expression(expr=2 * func)
        m.y.fix(3)
        e = m.z + m.x * m.E
        named_exprs, var_list, fixed_vars, external_funcs = collector(e, *args)
        self.assertEqual([m.E], named_exprs)
        self.assertEqual([m.z, m.x, m.y], var_list)
        self.assertEqual([m.y], fixed_vars)
        self.assertEqual([func], external_funcs)

    def test_external(self):
        self.external_func_helper(collect_vars_and_named_exprs)

    @unittest.skipUnless(cmodel_available, 'appsi extensions are not available')
    def test_external_cmodel(self):
        self.basics_helper(cmodel.prep_for_repn, cmodel.PyomoExprTypes())



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
"""Functions for solving the discrete problem."""

from copy import deepcopy

from pyomo.common.deprecation import deprecation_warning
from pyomo.common.errors import InfeasibleConstraintException
from pyomo.contrib.fbbt.fbbt import fbbt
from pyomo.contrib.gdpopt.util import (
    SuppressInfeasibleWarning,
    _DoNothing,
    get_main_elapsed_time,
)
from pyomo.core import Objective, Constraint
from pyomo.opt import SolutionStatus, SolverFactory
from pyomo.opt import TerminationCondition as tc
from pyomo.solvers.plugins.solvers.persistent_solver import PersistentSolver


def solve_MILP_discrete_problem(util_block, solver, config):
    """Solves the linear GDP model and attempts to resolve solution issues.
    Returns one of TerminationCondition.optimal, TerminationCondition.feasible,
    TerminationCondition.infeasible, or TerminationCondition.unbounded.
    """
    timing = solver.timing
    m = util_block.parent_block()

    if config.mip_presolve:
        try:
            # ESJ: deactivating satisfied constraints here is risky since it can
            # result in sending no constraints to the solver and hence not
            # actually filling in the values of the variables that got
            # implicitly fixed by these bounds. We can fix this by calling some
            # contrib.preprocessing transformations, but for now I'm just
            # leaving the constraints in.
            fbbt(
                m,
                integer_tol=config.integer_tolerance,
                deactivate_satisfied_constraints=False,
            )
            # [ESJ 1/28/22]: Despite being a little scary, the tightened bounds
            # are okay to leave in because if you tighten the bounds now, they
            # could only get tighter in later iterations, since you are
            # tightening this relaxation
        except InfeasibleConstraintException as e:
            config.logger.debug(
                "MIP preprocessing detected infeasibility:\n\t%s" % str(e)
            )
            return tc.infeasible

    # Deactivate extraneous IMPORT/EXPORT suffixes
    getattr(m, 'ipopt_zL_out', _DoNothing()).deactivate()
    getattr(m, 'ipopt_zU_out', _DoNothing()).deactivate()

    # Create solver, check availability
    if not SolverFactory(config.mip_solver).available():
        raise RuntimeError("MIP solver %s is not available." % config.mip_solver)

    # Callback immediately before solving MIP discrete problem
    config.call_before_discrete_problem_solve(solver, m, util_block)
    if config.call_before_master_solve is not _DoNothing:
        deprecation_warning(
            "The 'call_before_master_solve' argument is deprecated. "
            "Please use the 'call_before_discrete_problem_solve' option "
            "to specify the callback.",
            version="6.4.2",
        )

    with SuppressInfeasibleWarning():
        mip_args = dict(config.mip_solver_args)
        if config.time_limit is not None:
            elapsed = get_main_elapsed_time(timing)
            remaining = max(config.time_limit - elapsed, 1)
            if config.mip_solver == 'gams':
                mip_args['add_options'] = mip_args.get('add_options', [])
                mip_args['add_options'].append('option reslim=%s;' % remaining)
            elif config.mip_solver == 'multisolve':
                mip_args['time_limit'] = min(
                    mip_args.get('time_limit', float('inf')), remaining
                )
        results = SolverFactory(config.mip_solver).solve(m, **mip_args)

    config.call_after_discrete_problem_solve(solver, m, util_block)
    if config.call_after_master_solve is not _DoNothing:
        deprecation_warning(
            "The 'call_after_master_solve' argument is deprecated. "
            "Please use the 'call_after_discrete_problem_solve' option to "
            "specify the callback.",
            version="6.4.2",
        )

    terminate_cond = results.solver.termination_condition
    if terminate_cond is tc.infeasibleOrUnbounded:
        # Linear solvers will sometimes tell me that it's infeasible or
        # unbounded during presolve, but fails to distinguish. We need to
        # resolve with a solver option flag on.
        results, terminate_cond = distinguish_mip_infeasible_or_unbounded(m, config)
    if terminate_cond is tc.unbounded:
        # Solution is unbounded. This occurs when the objective is
        # nonlinear. The nonlinear objective is moved to the constraints, and
        # deactivated for the linear discrete problem. We will generate an
        # arbitrary discrete solution by bounding the objective and re-solving,
        # in hopes that the cuts we generate later bound this problem.

        obj_bound = 1e15
        config.logger.warning(
            'Discrete problem was unbounded. '
            'Re-solving with arbitrary bound values of (-{0:.10g}, {0:.10g}) '
            'on the objective, in order to get a discrete solution. '
            'Check your initialization routine.'.format(obj_bound)
        )
        discrete_objective = next(m.component_data_objects(Objective, active=True))
        util_block.objective_bound = Constraint(
            expr=(-obj_bound, discrete_objective.expr, obj_bound)
        )
        with SuppressInfeasibleWarning():
            results = SolverFactory(config.mip_solver).solve(
                m, **config.mip_solver_args
            )
        # get rid of the made-up constraint
        del util_block.objective_bound
        if results.solver.termination_condition in {
            tc.optimal,
            tc.feasible,
            tc.locallyOptimal,
            tc.globallyOptimal,
        }:
            # we found a solution, that's all we need to keep going.
            return tc.unbounded
        else:
            raise RuntimeError(
                "Unable to find a feasible solution for the "
                "unbounded MILP discrete problem by bounding "
                "the objective. Either check your "
                "discrete problem initialization, or add a "
                "bound on the discrete problem objective value "
                "that admits a feasible solution."
            )

    if terminate_cond is tc.optimal:
        return tc.optimal
    elif terminate_cond in {tc.locallyOptimal, tc.feasible}:
        return tc.feasible
    elif terminate_cond is tc.infeasible:
        config.logger.info(
            'MILP discrete problem is now infeasible. GDPopt has explored or '
            'cut off all feasible discrete configurations.'
        )
        return tc.infeasible
    elif terminate_cond is tc.maxTimeLimit:
        if len(results.solution) > 0:
            config.logger.info(
                'Unable to optimize MILP discrete problem within time limit. '
                'Using current solver feasible solution.'
            )
            return tc.feasible
        else:
            config.logger.info(
                'Unable to optimize MILP discrete problem within time limit. '
                'No solution found. Treating as infeasible, but there are no '
                'guarantees.'
            )
            return tc.infeasible
    elif (
        terminate_cond is tc.other
        and results.solution.status is SolutionStatus.feasible
    ):
        # load the solution and suppress the warning message by setting
        # solver status to ok.
        config.logger.info(
            'MIP solver reported feasible solution to MILP discrete problem, '
            'but it is not guaranteed to be optimal.'
        )
        return tc.feasible
    else:
        raise ValueError(
            'GDPopt unable to handle MILP discrete problem '
            'termination condition '
            'of %s. Solver message: %s' % (terminate_cond, results.solver.message)
        )


def distinguish_mip_infeasible_or_unbounded(m, config):
    """Distinguish between an infeasible or unbounded solution.

    Linear solvers will sometimes tell me that a problem is infeasible or
    unbounded during presolve, but not distinguish between the two cases. We
    address this by solving again with a solver option flag on.

    """
    tmp_args = deepcopy(config.mip_solver_args)
    if config.mip_solver == 'gurobi':
        # This solver option is specific to Gurobi.
        tmp_args['options'] = tmp_args.get('options', {})
        tmp_args['options']['DualReductions'] = 0
    mipopt = SolverFactory(config.mip_solver)
    # gdpopt no longer supports non-auto persistent solvers, but mindtpy does,
    # and it uses this function.
    if isinstance(mipopt, PersistentSolver):
        mipopt.set_instance(m)
    with SuppressInfeasibleWarning():
        results = mipopt.solve(m, load_solutions=False, **tmp_args)
        if len(results.solution) > 0:
            m.solutions.load_from(results)
    termination_condition = results.solver.termination_condition
    return results, termination_condition



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


from math import isnan
import unittest

from pyomo.core.expr import SumExpression, MonomialTermExpression
from pyomo.core.expr.compare import assertExpressionsEqual
from pyomo.environ import Any, ConcreteModel, log, Param, Var
from pyomo.repn.parameterized import ParameterizedQuadraticRepnVisitor
from pyomo.repn.tests.test_linear import VisitorConfig
from pyomo.repn.util import InvalidNumber

nan = float('nan')


def build_test_model():
    m = ConcreteModel()
    m.x = Var()
    m.y = Var()
    m.z = Var()
    m.p = Param(initialize=1, mutable=True)

    return m


class TestParameterizedQuadratic(unittest.TestCase):
    def test_constant_literal(self):
        """
        Ensure ParameterizedQuadraticRepnVisitor(*args, wrt=[]) works
        like QuadraticRepnVisitor.
        """
        expr = 2

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 2)
        self.assertEqual(repn.linear, {})
        self.assertIsNone(repn.quadratic)
        self.assertIsNone(repn.nonlinear)
        self.assertEqual(repn.to_expression(visitor), 2)

    def test_constant_param(self):
        m = build_test_model()
        m.p.set_value(2)
        expr = 2 + m.p

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, 4)
        self.assertEqual(repn.linear, {})
        self.assertIsNone(repn.quadratic)
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(self, repn.to_expression(visitor), 4)

    def test_binary_sum_identical_terms(self):
        m = build_test_model()
        expr = m.x + m.x

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y, m.z])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x): 2})
        self.assertIsNone(repn.quadratic)
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(self, repn.to_expression(visitor), 2 * m.x)

    def test_binary_sum_identical_terms_wrt_x(self):
        m = build_test_model()
        expr = m.x + m.x

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.x])
        # note: covers walker_exitNode for case where
        #       constant is a fixed expression
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, m.x + m.x)
        self.assertEqual(repn.linear, {})
        self.assertIsNone(repn.quadratic)
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(self, repn.to_expression(visitor), m.x + m.x)

    def test_binary_sum_nonidentical_terms(self):
        m = build_test_model()
        expr = m.x + m.y

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x): 1, id(m.y): 1})
        self.assertIsNone(repn.quadratic)
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(self, repn.to_expression(visitor), m.x + m.y)

    def test_binary_sum_nonidentical_terms_wrt_x(self):
        m = build_test_model()
        expr = m.x + m.y

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.x])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.y): 0})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, m.x)
        self.assertEqual(repn.linear, {id(m.y): 1})
        self.assertIsNone(repn.quadratic)
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(self, repn.to_expression(visitor), m.y + m.x)

    def test_ternary_sum_with_product(self):
        m = build_test_model()
        e = m.x + m.z * m.y + m.z

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        repn = visitor.walk_expression(e)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.z): m.z, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.z): 1, id(m.y): 2})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(len(repn.linear), 2)
        self.assertEqual(repn.linear[id(m.x)], 1)
        self.assertEqual(repn.linear[id(m.z)], 1)
        self.assertEqual(len(repn.quadratic), 1)
        self.assertEqual(repn.quadratic[(id(m.z), id(m.y))], 1)
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(self, repn.to_expression(visitor), m.z * m.y + m.x + m.z)

    def test_ternary_sum_with_product_wrt_z(self):
        m = build_test_model()
        e = m.x + m.z * m.y + m.z

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.z])
        repn = visitor.walk_expression(e)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertIs(repn.constant, m.z)
        self.assertEqual(len(repn.linear), 2)
        self.assertEqual(repn.linear[id(m.x)], 1)
        self.assertIs(repn.linear[id(m.y)], m.z)
        self.assertIsNone(repn.quadratic)
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(self, repn.to_expression(visitor), m.x + m.z * m.y + m.z)

    def test_nonlinear_wrt_x(self):
        m = build_test_model()
        expr = log(m.x)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.x])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, log(m.x))
        self.assertEqual(repn.linear, {})
        self.assertIsNone(repn.quadratic)
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(self, repn.to_expression(visitor), log(m.x))

    def test_linear_constant_coeffs(self):
        m = build_test_model()
        e = 2 + 3 * m.x

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(e)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 2)
        self.assertEqual(repn.linear, {id(m.x): 3})
        self.assertIsNone(repn.quadratic)
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(self, repn.to_expression(visitor), 3 * m.x + 2)

    def test_linear_constant_coeffs_wrt_x(self):
        m = build_test_model()
        e = 2 + 3 * m.x

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.x])
        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(e)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, 2 + 3 * m.x)
        self.assertEqual(repn.linear, {})
        self.assertIsNone(repn.quadratic)
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(self, repn.to_expression(visitor), 2 + 3 * m.x)

    def test_quadratic(self):
        m = build_test_model()
        e = 2 + 3 * m.x + 4 * m.x**2

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(e)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 2)
        self.assertEqual(repn.linear, {id(m.x): 3})
        self.assertEqual(repn.quadratic, {(id(m.x), id(m.x)): 4})
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(
            self, repn.to_expression(visitor), 4 * m.x**2 + 3 * m.x + 2
        )

    def test_product_quadratic_quadratic(self):
        m = build_test_model()
        e = (2 + 3 * m.x + 4 * m.x**2) * (5 + 6 * m.x + 7 * m.x**2)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(e)

        QE4 = 4 * m.x**2
        QE7 = 7 * m.x**2
        LE3 = MonomialTermExpression((3, m.x))
        LE6 = MonomialTermExpression((6, m.x))
        NL = +QE4 * (QE7 + LE6) + (LE3) * (QE7)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 10)
        self.assertEqual(repn.linear, {id(m.x): 27})
        self.assertEqual(repn.quadratic, {(id(m.x), id(m.x)): 52})
        assertExpressionsEqual(self, repn.nonlinear, NL)
        assertExpressionsEqual(
            self, repn.to_expression(visitor), NL + 52 * m.x**2 + 27 * m.x + 10
        )

    def test_product_quadratic_quadratic_2(self):
        m = build_test_model()
        e = (2 + 3 * m.x + 4 * m.x**2) * (5 + 6 * m.x + 7 * m.x**2)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        visitor.expand_nonlinear_products = False
        repn = visitor.walk_expression(e)

        NL = (4 * m.x**2 + 3 * m.x + 2) * (7 * m.x**2 + 6 * m.x + 5)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertIsNone(repn.quadratic)
        assertExpressionsEqual(self, repn.nonlinear, NL)
        assertExpressionsEqual(self, repn.to_expression(visitor), NL)

    def test_product_linear_linear(self):
        m = build_test_model()
        e = (1 + 2 * m.x + 3 * m.y) * (4 + 5 * m.x + 6 * m.y)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        repn = visitor.walk_expression(e)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 4)
        self.assertEqual(repn.linear, {id(m.x): 13, id(m.y): 18})
        self.assertEqual(
            repn.quadratic,
            {(id(m.x), id(m.x)): 10, (id(m.y), id(m.y)): 18, (id(m.x), id(m.y)): 27},
        )
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(
            self,
            repn.to_expression(visitor),
            (10 * m.x**2 + 27 * (m.x * m.y) + 18 * m.y**2 + 13 * m.x + 18 * m.y + 4),
        )

    def test_product_linear_linear_wrt_y(self):
        m = build_test_model()
        e = (1 + 2 * m.x + 3 * m.y) * (4 + 5 * m.x + 6 * m.y)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y, m.z])
        repn = visitor.walk_expression(e)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, (1 + 3 * m.y) * (4 + 6 * m.y))
        self.assertEqual(len(repn.linear), 1)
        assertExpressionsEqual(
            self, repn.linear[id(m.x)], (4 + 6 * m.y) * 2 + (1 + 3 * m.y) * 5
        )
        self.assertEqual(repn.quadratic, {(id(m.x), id(m.x)): 10})
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(
            self,
            repn.to_expression(visitor),
            (
                10 * m.x**2
                + ((4 + 6 * m.y) * 2 + (1 + 3 * m.y) * 5) * m.x
                + (1 + 3 * m.y) * (4 + 6 * m.y)
            ),
        )

    def test_product_linear_linear_const_0(self):
        m = build_test_model()
        expr = (0 + 3 * m.x + 4 * m.y) * (5 + 3 * m.x + 7 * m.y)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x): 15, id(m.y): 20})
        self.assertEqual(
            repn.quadratic,
            {(id(m.x), id(m.x)): 9, (id(m.x), id(m.y)): 33, (id(m.y), id(m.y)): 28},
        )
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(
            self,
            repn.to_expression(visitor),
            9 * m.x**2 + 33 * (m.x * m.y) + 28 * m.y**2 + 15 * m.x + 20 * m.y,
        )

    def test_product_linear_quadratic(self):
        m = build_test_model()
        expr = (5 + 3 * m.x + 7 * m.y) * (1 + 3 * m.x + 4 * m.y + 8 * m.y * m.x)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        repn = visitor.walk_expression(expr)

        NL = (3 * m.x + 7 * m.y + 5) * (8 * (m.x * m.y) + 3 * m.x + 4 * m.y + 1)

        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertIsNone(repn.quadratic)
        assertExpressionsEqual(self, repn.nonlinear, NL)
        assertExpressionsEqual(self, repn.to_expression(visitor), NL)

        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 5)
        self.assertEqual(repn.linear, {id(m.x): 18, id(m.y): 27})
        self.assertEqual(
            repn.quadratic,
            {(id(m.x), id(m.y)): 73, (id(m.x), id(m.x)): 9, (id(m.y), id(m.y)): 28},
        )
        assertExpressionsEqual(
            self, repn.nonlinear, (3 * m.x + 7 * m.y) * (8 * (m.x * m.y))
        )
        assertExpressionsEqual(
            self,
            repn.to_expression(visitor),
            (
                73 * (m.x * m.y)
                + 9 * m.x**2
                + 28 * m.y**2
                + 18 * m.x
                + 27 * m.y
                + 5
                + (3 * m.x + 7 * m.y) * (8 * (m.x * m.y))
            ),
        )

    def test_product_linear_quadratic_wrt_x(self):
        m = build_test_model()
        expr = (0 + 3 * m.x + 4 * m.y + 8 * m.y * m.x) * (5 + 3 * m.x + 7 * m.y)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.x])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.y): 0})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, 3 * m.x * (5 + 3 * m.x))
        self.assertEqual(len(repn.linear), 1)
        assertExpressionsEqual(
            self, repn.linear[id(m.y)], (5 + 3 * m.x) * (4 + 8 * m.x) + 21 * m.x
        )
        self.assertEqual(len(repn.quadratic), 1)
        assertExpressionsEqual(
            self, repn.quadratic[id(m.y), id(m.y)], (4 + 8 * m.x) * 7
        )
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(
            self,
            repn.to_expression(visitor),
            (4 + 8 * m.x) * 7 * m.y**2
            + ((5 + 3 * m.x) * (4 + 8 * m.x) + 21 * m.x) * m.y
            + 3 * m.x * (5 + 3 * m.x),
        )

    def test_product_nonlinear_var_expand_false(self):
        m = build_test_model()
        e = (m.x + m.y + log(m.x)) * m.x

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        visitor.expand_nonlinear_products = False
        repn = visitor.walk_expression(e)

        NL = (log(m.x) + (m.x + m.y)) * m.x

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertIsNone(repn.quadratic)
        assertExpressionsEqual(self, repn.nonlinear, NL)
        assertExpressionsEqual(self, repn.to_expression(visitor), NL)

    def test_product_nonlinear_var_expand_true(self):
        m = build_test_model()
        e = (m.x + m.y + log(m.x)) * m.x

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(e)

        NL = log(m.x) * m.x

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.quadratic, {(id(m.x), id(m.x)): 1, (id(m.x), id(m.y)): 1})
        assertExpressionsEqual(self, repn.nonlinear, NL)

    def test_product_nonlinear_var_2_expand_false(self):
        m = build_test_model()
        e = m.x * (m.x + m.y + log(m.x) + 2)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        visitor.expand_nonlinear_products = False
        repn = visitor.walk_expression(e)

        NL = m.x * (log(m.x) + (m.x + m.y + 2))

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertIsNone(repn.quadratic)
        assertExpressionsEqual(self, repn.nonlinear, NL)
        assertExpressionsEqual(self, repn.to_expression(visitor), NL)

    def test_product_nonlinear_var_2_expand_true(self):
        m = build_test_model()
        e = m.x * (m.x + m.y + log(m.x) + 2)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(e)

        NL = m.x * log(m.x)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x): 2})
        self.assertEqual(repn.quadratic, {(id(m.x), id(m.x)): 1, (id(m.x), id(m.y)): 1})
        assertExpressionsEqual(self, repn.nonlinear, NL)
        assertExpressionsEqual(
            self, repn.to_expression(visitor), m.x**2 + m.x * m.y + 2 * m.x + NL
        )

    def test_zero_elimination(self):
        m = ConcreteModel()
        m.x = Var(range(4))
        e = 0 * m.x[0] + 0 * m.x[1] * m.x[2] + 0 * log(m.x[3])

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        repn = visitor.walk_expression(e)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(
            cfg.var_map,
            {
                id(m.x[0]): m.x[0],
                id(m.x[1]): m.x[1],
                id(m.x[2]): m.x[2],
                id(m.x[3]): m.x[3],
            },
        )
        self.assertEqual(
            cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2, id(m.x[3]): 3}
        )
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertIsNone(repn.quadratic)
        assertExpressionsEqual(self, repn.nonlinear, 0 * log(m.x[3]))
        assertExpressionsEqual(self, repn.to_expression(visitor), 0 * log(m.x[3]))

    def test_uninitialized_param_expansion(self):
        m = ConcreteModel()
        m.x = Var(range(4))
        m.p = Param(mutable=True, within=Any, initialize=None)
        e = m.p * m.x[0] + m.p * m.x[1] * m.x[2] + m.p * log(m.x[3])

        cfg = VisitorConfig()
        repn = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[]).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(
            cfg.var_map,
            {
                id(m.x[0]): m.x[0],
                id(m.x[1]): m.x[1],
                id(m.x[2]): m.x[2],
                id(m.x[3]): m.x[3],
            },
        )
        self.assertEqual(
            cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2, id(m.x[3]): 3}
        )
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x[0]): InvalidNumber(None)})
        self.assertEqual(
            repn.quadratic, {(id(m.x[1]), id(m.x[2])): InvalidNumber(None)}
        )
        self.assertEqual(repn.nonlinear, InvalidNumber(None))

    def test_zero_times_var(self):
        m = build_test_model()
        e = 0 * m.x

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        repn = visitor.walk_expression(e)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertIsNone(repn.quadratic)
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(self, repn.to_expression(visitor), 0)

    def test_square_linear(self):
        m = build_test_model()
        expr = (1 + 3 * m.x + 4 * m.y) ** 2

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 1)
        self.assertEqual(repn.linear, {id(m.x): 6, id(m.y): 8})
        self.assertEqual(
            repn.quadratic,
            {(id(m.x), id(m.x)): 9, (id(m.y), id(m.y)): 16, (id(m.x), id(m.y)): 24},
        )
        self.assertEqual(repn.nonlinear, None)
        assertExpressionsEqual(
            self,
            repn.to_expression(visitor),
            9 * m.x**2 + 16 * m.y**2 + 24 * (m.x * m.y) + 6 * m.x + 8 * m.y + 1,
        )

    def test_square_linear_wrt_y(self):
        m = build_test_model()
        expr = (1 + 3 * m.x + 4 * m.y) ** 2

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y, m.z])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, (1 + 4 * m.y) ** 2)
        self.assertEqual(len(repn.linear), 1)
        assertExpressionsEqual(self, repn.linear[id(m.x)], 3 * (2 * (1 + 4 * m.y)))
        self.assertEqual(repn.quadratic, {(id(m.x), id(m.x)): 9})
        self.assertEqual(repn.nonlinear, None)
        assertExpressionsEqual(
            self,
            repn.to_expression(visitor),
            (9 * m.x**2 + 3 * (2 * (1 + 4 * m.y)) * m.x + (1 + 4 * m.y) ** 2),
        )

    def test_square_linear_float(self):
        m = build_test_model()
        expr = (1 + 3 * m.x + 4 * m.y) ** 2.0

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 1)
        self.assertEqual(repn.linear, {id(m.x): 6, id(m.y): 8})
        self.assertEqual(
            repn.quadratic,
            {(id(m.x), id(m.x)): 9, (id(m.y), id(m.y)): 16, (id(m.x), id(m.y)): 24},
        )
        self.assertEqual(repn.nonlinear, None)
        assertExpressionsEqual(
            self,
            repn.to_expression(visitor),
            9 * m.x**2 + 16 * m.y**2 + 24 * (m.x * m.y) + 6 * m.x + 8 * m.y + 1,
        )

    def test_division_quadratic_nonlinear(self):
        m = build_test_model()
        expr = (1 + 3 * m.x + 4 * log(m.x) * m.y + 4 * m.y**2) / (2 * m.x)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertIsNone(repn.quadratic)
        assertExpressionsEqual(
            self,
            repn.nonlinear,
            (4 * m.y**2 + 3 * m.x + 1 + log(m.x) * 4 * m.y) / (2 * m.x),
        )
        assertExpressionsEqual(self, repn.to_expression(visitor), repn.nonlinear)

    def test_division_quadratic_nonlinear_wrt_x(self):
        m = build_test_model()
        expr = (1 + 3 * m.x + 4 * log(m.x) * m.y + 4 * m.y**2) / (2 * m.x)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.x])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.y): 0})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, (1 + 3 * m.x) * (1 / (2 * m.x)))
        self.assertEqual(len(repn.linear), 1)
        assertExpressionsEqual(
            self, repn.linear[id(m.y)], (4 * log(m.x)) * (1 / (2 * m.x))
        )
        self.assertEqual(len(repn.quadratic), 1)
        assertExpressionsEqual(
            self, repn.quadratic[id(m.y), id(m.y)], 4 * (1 / (2 * m.x))
        )
        self.assertEqual(repn.nonlinear, None)
        assertExpressionsEqual(
            self,
            repn.to_expression(visitor),
            (4 * (1 / (2 * m.x))) * m.y**2
            + ((4 * log(m.x)) * (1 / (2 * m.x))) * m.y
            + (1 + 3 * m.x) * (1 / (2 * m.x)),
        )

    def test_constant_expr_multiplier(self):
        m = build_test_model()
        expr = 5 * (2 * m.x + m.x**2)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x): 10})
        self.assertEqual(repn.quadratic, {(id(m.x), id(m.x)): 5})
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(self, repn.to_expression(visitor), 5 * m.x**2 + 10 * m.x)

    def test_0_mult_nan_linear_coeff(self):
        m = build_test_model()
        expr = 0 * (nan * m.x + m.y + log(m.x) + m.y * m.x**2 + 2 * m.x)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, 0 * m.y)
        self.assertEqual(len(repn.linear), 1)
        assertExpressionsEqual(self, repn.linear[id(m.x)], nan)
        self.assertEqual(len(repn.quadratic), 1)
        assertExpressionsEqual(self, repn.quadratic[id(m.x), id(m.x)], 0 * m.y)
        assertExpressionsEqual(self, repn.nonlinear, (log(m.x)) * 0)
        assertExpressionsEqual(
            self,
            repn.to_expression(visitor),
            0 * m.y * m.x**2 + nan * m.x + 0 * m.y + log(m.x) * 0,
        )

    def test_0_mult_nan_quadratic_coeff(self):
        m = build_test_model()
        expr = 0 * (m.x + m.y + log(m.x) + nan * m.x**2 + 2 * m.x)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y])
        # visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, 0 * m.y)
        self.assertEqual(repn.linear, {})
        self.assertEqual(len(repn.quadratic), 1)
        assertExpressionsEqual(self, repn.quadratic[id(m.x), id(m.x)], nan)
        assertExpressionsEqual(self, repn.nonlinear, (log(m.x)) * 0)
        assertExpressionsEqual(
            self, repn.to_expression(visitor), nan * m.x**2 + 0 * m.y + log(m.x) * 0
        )

    def test_square_quadratic(self):
        m = build_test_model()
        expr = (1 + m.x + m.y + m.x**2 + m.x * m.y) ** 2.0

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        repn = visitor.walk_expression(expr)

        NL = m.x**2 + m.x * m.y + m.x + m.y + 1

        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, 0)
        self.assertEqual(len(repn.linear), 0)
        self.assertIsNone(repn.quadratic)
        assertExpressionsEqual(self, repn.nonlinear, NL * NL)
        assertExpressionsEqual(self, repn.to_expression(visitor), NL * NL)

        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(expr)

        NL = (m.x**2 + m.x * m.y) * (m.x**2 + m.x * m.y + m.x + m.y) + (
            m.x + m.y
        ) * (m.x**2 + m.x * m.y)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 1)
        self.assertEqual(repn.linear, {id(m.x): 2, id(m.y): 2})
        self.assertEqual(
            repn.quadratic,
            {(id(m.x), id(m.x)): 3, (id(m.x), id(m.y)): 4, (id(m.y), id(m.y)): 1},
        )
        assertExpressionsEqual(self, repn.nonlinear, NL)
        assertExpressionsEqual(
            self,
            repn.to_expression(visitor),
            NL + 3 * m.x**2 + 4 * (m.x * m.y) + m.y**2 + 2 * m.x + 2 * m.y + 1,
        )

    def test_square_quadratic_wrt_y(self):
        m = build_test_model()
        expr = (1 + m.x + m.y + m.x**2 + m.x * m.y) ** 2.0

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y])
        repn = visitor.walk_expression(expr)

        NL = m.x**2 + (1 + m.y) * m.x + (1 + m.y)

        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, 0)
        self.assertEqual(len(repn.linear), 0)
        self.assertIsNone(repn.quadratic)
        assertExpressionsEqual(self, repn.nonlinear, NL * NL)
        assertExpressionsEqual(self, repn.to_expression(visitor), NL * NL)

        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(expr)

        NL = m.x**2 * (m.x**2 + (1 + m.y) * m.x) + ((1 + m.y) * m.x) * m.x**2
        QC = 1 + m.y + 1 + m.y + (1 + m.y) * (1 + m.y)
        LC = (1 + m.y) * (1 + m.y) + (1 + m.y) * (1 + m.y)
        CON = (1 + m.y) * (1 + m.y)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, (1 + m.y) * (1 + m.y))
        self.assertEqual(len(repn.linear), 1)
        assertExpressionsEqual(
            self, repn.linear[id(m.x)], (1 + m.y) * (1 + m.y) + (1 + m.y) * (1 + m.y)
        )
        self.assertEqual(len(repn.quadratic), 1)
        assertExpressionsEqual(
            self,
            repn.quadratic[id(m.x), id(m.x)],
            1 + m.y + 1 + m.y + (1 + m.y) * (1 + m.y),
        )
        assertExpressionsEqual(self, repn.nonlinear, NL)
        assertExpressionsEqual(
            self, repn.to_expression(visitor), NL + QC * m.x**2 + LC * m.x + CON
        )

    def test_cube_linear(self):
        m = build_test_model()
        expr = (1 + m.x + m.y) ** 3

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertIsNone(repn.quadratic)
        # cubic expansion not supported
        assertExpressionsEqual(self, repn.nonlinear, (m.x + m.y + 1) ** 3)
        assertExpressionsEqual(self, repn.to_expression(visitor), (m.x + m.y + 1) ** 3)

    def test_nonlinear_product_with_constant_terms(self):
        m = build_test_model()
        # test product of nonlinear expressions where one
        # multiplicand has constant of value 1
        expr = (1 + log(m.x)) * (log(m.x) + m.y**2)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.z])
        repn = visitor.walk_expression(expr)
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertIsNone(repn.quadratic)
        assertExpressionsEqual(
            self, repn.nonlinear, (log(m.x) + 1) * (log(m.x) + m.y**2)
        )
        assertExpressionsEqual(
            self, repn.to_expression(visitor), (log(m.x) + 1) * (log(m.x) + m.y**2)
        )

        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.quadratic, {(id(m.y), id(m.y)): 1})
        assertExpressionsEqual(
            self, repn.nonlinear, log(m.x) * (log(m.x) + m.y**2) + log(m.x)
        )
        assertExpressionsEqual(
            self,
            repn.to_expression(visitor),
            log(m.x) * (log(m.x) + m.y**2) + log(m.x) + m.y**2,
        )

    def test_finalize_simplify_coefficients(self):
        m = build_test_model()
        expr = m.x + m.p * m.x**2 + 2 * m.y**2 - m.x - m.p * m.x**2 - m.p * m.z

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.z): m.z})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.z): 1})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, 2 * m.y**2)
        self.assertEqual(repn.linear, {id(m.z): -1})
        self.assertEqual(repn.quadratic, None)
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(self, repn.to_expression(visitor), -1 * m.z + 2 * m.y**2)

    def test_factor_multiplier_simplify_coefficients(self):
        m = build_test_model()
        expr = 2 * (m.x + m.x**2 + 2 * m.y**2 - m.x - m.x**2 - m.p * m.z)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y])
        # this tests case where there are zeros in the `linear`
        # and `quadratic` dicts of the unfinalized repn
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.z): m.z})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.z): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertIsNone(repn.nonlinear)
        self.assertEqual(repn.quadratic, None)
        self.assertEqual(repn.linear, {id(m.z): -2})
        assertExpressionsEqual(self, repn.constant, (2 * m.y**2) * 2)
        assertExpressionsEqual(
            self, repn.to_expression(visitor), -2 * m.z + (2 * m.y**2) * 2
        )

    def test_sum_nonlinear_custom_multiplier(self):
        m = build_test_model()
        expr = 2 * (1 + log(m.x)) + (2 * (m.y + m.y**2 + log(m.x)))

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, 2 + 2 * (m.y + m.y**2))
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.quadratic, None)
        assertExpressionsEqual(self, repn.nonlinear, 2 * log(m.x) + 2 * log(m.x))
        assertExpressionsEqual(
            self,
            repn.to_expression(visitor),
            2 * log(m.x) + 2 * log(m.x) + 2 + 2 * (m.y + m.y**2),
        )

    def test_negation_linear(self):
        m = build_test_model()
        expr = -(2 + 3 * m.x + 5 * m.x * m.y)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, -2)
        self.assertEqual(len(repn.linear), 1)
        assertExpressionsEqual(self, repn.linear[id(m.x)], (3 + 5 * m.y) * -1)
        self.assertIsNone(repn.quadratic)
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(
            self, repn.to_expression(visitor), (3 + 5 * m.y) * -1 * m.x - 2
        )

    def test_negation_nonlinear_wrt_y_fix_z(self):
        m = build_test_model()
        m.z.fix(2)
        expr = -(
            2
            + 3 * m.x
            + 4 * m.y * m.z
            + 5 * m.x**2 * m.y
            + 6 * m.x * (m.z - 2)
            + m.z**2
            + m.z * log(m.x)
        )

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, (2 + 8 * m.y + 4) * -1)
        self.assertEqual(repn.linear, {id(m.x): -3})
        self.assertEqual(len(repn.quadratic), 1)
        assertExpressionsEqual(self, repn.quadratic[(id(m.x), id(m.x))], -5 * m.y)
        assertExpressionsEqual(self, repn.nonlinear, 2 * log(m.x) * -1)
        assertExpressionsEqual(
            self,
            repn.to_expression(visitor),
            +(-5 * m.y) * (m.x**2)
            + (-3) * m.x
            + (2 + 8 * m.y + 4) * (-1)
            + 2 * log(m.x) * -1,
        )

    def test_negation_product_linear_linear(self):
        m = build_test_model()
        expr = -(1 + 2 * m.x + 3 * m.y) * (4 + 5 * m.x + 6 * m.y * 7 * m.z)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y, m.z])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(
            self, repn.constant, (1 + 3 * m.y) * (4 + 42 * m.y * m.z) * (-1)
        )
        self.assertEqual(len(repn.linear), 1)
        assertExpressionsEqual(
            self,
            repn.linear[id(m.x)],
            ((4 + 42 * m.y * m.z) * 2 + (1 + 3 * m.y) * 5) * -1,
        )
        self.assertEqual(len(repn.quadratic), 1)
        assertExpressionsEqual(self, repn.quadratic[id(m.x), id(m.x)], -10)
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(
            self,
            repn.to_expression(visitor),
            (
                -10 * m.x**2
                + ((4 + 42 * m.y * m.z) * 2 + (1 + 3 * m.y) * 5) * -1 * m.x
                + (1 + 3 * m.y) * (4 + 42 * m.y * m.z) * (-1)
            ),
        )

    def test_expanded_monomial_square_term(self):
        m = build_test_model()
        expr = m.x * m.x * m.p

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.z])
        # ensure overcomplication issues with standard repn
        # are not repeated by quadratic repn
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.quadratic, {(id(m.x), id(m.x)): 1})
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(self, repn.to_expression(visitor), m.x**2)

    def test_sum_bilinear_terms_commute_product(self):
        m = build_test_model()
        expr = m.x * m.y + m.y * m.x

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.z])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.quadratic, {(id(m.x), id(m.y)): 2})
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(self, repn.to_expression(visitor), 2 * (m.x * m.y))

    def test_sum_nonlinear(self):
        m = build_test_model()
        expr = (1 + log(m.x)) + (m.x + m.y + m.y**2 + log(m.x))

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y, m.z])
        # tests special case of `repn.append` where multiplier
        # is 1 and both summands have a nonlinear term
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, 1 + m.y + m.y**2)
        self.assertEqual(repn.linear, {id(m.x): 1})
        self.assertIsNone(repn.quadratic)
        assertExpressionsEqual(self, repn.nonlinear, log(m.x) + log(m.x))
        assertExpressionsEqual(
            self,
            repn.to_expression(visitor),
            log(m.x) + log(m.x) + m.x + (1 + m.y) + m.y**2,
        )

    def test_product_linear_linear_0_nan(self):
        m = build_test_model()
        m.p.set_value(0)
        expr = (m.p + 0 * m.x) * (nan + nan * m.x)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y, m.z])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(len(repn.linear), 1)
        self.assertTrue(isnan(repn.linear[id(m.x)]))
        self.assertEqual(len(repn.quadratic), 1)
        self.assertTrue(isnan(repn.quadratic[id(m.x), id(m.x)]))
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(
            self, repn.to_expression(visitor), nan * m.x**2 + nan * m.x
        )

    def test_product_quadratic_quadratic_nan_0(self):

        m = build_test_model()
        m.p.set_value(0)
        expr = (nan + nan * m.x + nan * m.x**2) * (m.p + 0 * m.x + 0 * m.x**2)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y, m.z])
        repn = visitor.walk_expression(expr)

        NL = (nan * m.x**2 + nan * m.x + nan) * (0 * m.x**2 + 0 * m.x)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(len(repn.linear), 0)
        self.assertIsNone(repn.quadratic)
        assertExpressionsEqual(self, repn.nonlinear, NL)
        assertExpressionsEqual(self, repn.to_expression(visitor), NL)

        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(expr)

        NL = (nan * m.x**2) * (0 * m.x**2 + 0 * m.x) + nan * m.x * (0 * m.x**2)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertTrue(isnan(repn.constant))
        self.assertEqual(len(repn.linear), 1)
        self.assertTrue(isnan(repn.linear[id(m.x)]))
        self.assertEqual(len(repn.quadratic), 1)
        self.assertTrue(isnan(repn.quadratic[id(m.x), id(m.x)]))
        assertExpressionsEqual(self, repn.nonlinear, NL)
        assertExpressionsEqual(
            self, repn.to_expression(visitor), NL + nan * m.x**2 + nan * m.x + nan
        )

    def test_product_quadratic_quadratic_0_nan(self):
        m = build_test_model()
        m.p.set_value(0)
        expr = (m.p + 0 * m.x + 0 * m.x**2) * (nan + nan * m.x + nan * m.x**2)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y, m.z])
        repn = visitor.walk_expression(expr)

        NL = (0 * m.x**2 + 0 * m.x) * (nan * m.x**2 + nan * m.x + nan)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(len(repn.linear), 0)
        self.assertIsNone(repn.quadratic)
        assertExpressionsEqual(self, repn.nonlinear, NL)
        assertExpressionsEqual(self, repn.to_expression(visitor), NL)

        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(expr)

        NL = (0 * m.x**2) * (nan * m.x**2 + nan * m.x) + 0 * m.x * (nan * m.x**2)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertTrue(isnan(repn.constant))
        self.assertEqual(len(repn.linear), 1)
        self.assertTrue(isnan(repn.linear[id(m.x)]))
        self.assertEqual(len(repn.quadratic), 1)
        self.assertTrue(isnan(repn.quadratic[id(m.x), id(m.x)]))
        assertExpressionsEqual(self, repn.nonlinear, NL)
        assertExpressionsEqual(
            self, repn.to_expression(visitor), NL + nan * m.x**2 + nan * m.x + nan
        )

    def test_nary_sum_products(self):
        m = build_test_model()
        expr = (
            m.x**2 * (m.z - 1)
            + m.x * (m.y**4 + 0.8)
            - 5 * m.x * m.y * m.z
            + m.x * (m.y + 2)
        )

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y, m.z])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(len(repn.linear), 1)
        assertExpressionsEqual(
            self, repn.linear[id(m.x)], m.y**4 + 0.8 + 5 * m.y * m.z * (-1) + (m.y + 2)
        )
        assertExpressionsEqual(self, repn.quadratic[id(m.x), id(m.x)], m.z - 1)
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(
            self,
            repn.to_expression(visitor),
            (m.z - 1) * m.x**2
            + (m.y**4 + 0.8 + 5 * m.y * m.z * (-1) + (m.y + 2)) * m.x,
        )

    def test_ternary_product_linear(self):
        m = build_test_model()
        expr = (1 + 2 * m.x) * (3 + 4 * m.y) * (5 + 6 * m.z)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.z): m.z})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.z): 1})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, 5 * (3 + 4 * m.y))
        self.assertEqual(len(repn.linear), 2)
        assertExpressionsEqual(self, repn.linear[id(m.x)], 10 * (3 + 4 * m.y))
        assertExpressionsEqual(self, repn.linear[id(m.z)], 6 * (3 + 4 * m.y))
        self.assertEqual(len(repn.quadratic), 1)
        assertExpressionsEqual(
            self, repn.quadratic[id(m.x), id(m.z)], 12 * (3 + 4 * m.y)
        )
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(
            self,
            repn.to_expression(visitor),
            (
                12 * (3 + 4 * m.y) * (m.x * m.z)
                + 10 * (3 + 4 * m.y) * m.x
                + 6 * (3 + 4 * m.y) * m.z
                + 5 * (3 + 4 * m.y)
            ),
        )

    def test_noninteger_pow_linear(self):
        m = build_test_model()
        expr = (1 + 2 * m.x + 3 * m.y) ** 1.5

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y, m.z])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertIsNone(repn.quadratic)
        assertExpressionsEqual(self, repn.nonlinear, (2 * m.x + 1 + 3 * m.y) ** 1.5)
        assertExpressionsEqual(
            self, repn.to_expression(visitor), (2 * m.x + 1 + 3 * m.y) ** 1.5
        )

    def test_variable_pow_linear(self):
        m = build_test_model()
        expr = (1 + 2 * m.x + 3 * m.y) ** (m.y)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y, m.z])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertIsNone(repn.quadratic)
        assertExpressionsEqual(self, repn.nonlinear, (2 * m.x + 1 + 3 * m.y) ** m.y)
        assertExpressionsEqual(
            self, repn.to_expression(visitor), (2 * m.x + 1 + 3 * m.y) ** m.y
        )

    def test_pow_integer_fixed_var(self):
        m = build_test_model()
        m.z.fix(2)
        expr = (1 + 2 * m.x + 3 * m.y) ** (m.z)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, (1 + 3 * m.y) ** 2)
        self.assertEqual(len(repn.linear), 1)
        assertExpressionsEqual(self, repn.linear[id(m.x)], 2 * (2 * (1 + 3 * m.y)))
        self.assertEqual(repn.quadratic, {(id(m.x), id(m.x)): 4})
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(
            self,
            repn.to_expression(visitor),
            (4 * m.x**2 + (2 * (2 * (1 + 3 * m.y))) * m.x + (1 + 3 * m.y) ** 2),
        )

    def test_repr_parameterized_quadratic_repn(self):
        m = build_test_model()
        expr = 2 + m.x + m.x**2 + log(m.x)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y, m.z])
        repn = visitor.walk_expression(expr)

        linear_dict = {id(m.x): 1}
        quad_dict = {(id(m.x), id(m.x)): 1}
        expected_repn_str = (
            "ParameterizedQuadraticRepn("
            "mult=1, "
            "const=2, "
            f"linear={linear_dict}, "
            f"quadratic={quad_dict}, "
            "nonlinear=log(x))"
        )
        self.assertEqual(repr(repn), expected_repn_str)
        self.assertEqual(str(repn), expected_repn_str)

    def test_product_var_linear_wrt_yz(self):
        """
        Test product of Var and quadratic expression.

        Aimed at testing what happens when one multiplicand
        of a product
        has a constant term of 0, and the other has a
        constant term that is an expression.
        """
        m = build_test_model()
        expr = m.x * (m.y + m.x * m.y + m.z)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y, m.z])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, 0)
        self.assertEqual(len(repn.linear), 1)
        assertExpressionsEqual(self, repn.linear[id(m.x)], m.y + m.z)
        self.assertEqual(len(repn.quadratic), 1)
        assertExpressionsEqual(self, repn.quadratic[id(m.x), id(m.x)], m.y)
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(
            self, repn.to_expression(visitor), m.y * m.x**2 + (m.y + m.z) * m.x
        )

    def test_product_linear_var_wrt_yz(self):
        """
        Test product of Var and quadratic expression.

        Checks what happens when multiplicands of
        `test_product_var_linear` are swapped/commuted.
        """
        m = build_test_model()
        expr = (m.y + m.x * m.y + m.z) * m.x

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.y, m.z])
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, 0)
        self.assertEqual(len(repn.linear), 1)
        assertExpressionsEqual(self, repn.linear[id(m.x)], m.y + m.z)
        self.assertEqual(len(repn.quadratic), 1)
        assertExpressionsEqual(self, repn.quadratic[id(m.x), id(m.x)], m.y)
        self.assertIsNone(repn.nonlinear)
        assertExpressionsEqual(
            self, repn.to_expression(visitor), m.y * m.x**2 + (m.y + m.z) * m.x
        )

    def test_product_var_quadratic(self):
        """
        Test product of Var and quadratic expression.

        Aimed at testing what happens when one multiplicand
        of a product
        has a constant term of 0, and the other has a
        constant term that is an expression.
        """
        m = build_test_model()
        expr = m.x * (m.y + m.x * m.y + m.z)

        cfg = VisitorConfig()
        visitor = ParameterizedQuadraticRepnVisitor(**cfg, wrt=[m.z])
        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(expr)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        assertExpressionsEqual(self, repn.constant, 0 * m.z)
        self.assertEqual(len(repn.linear), 1)
        assertExpressionsEqual(self, repn.linear[id(m.x)], m.z)
        self.assertEqual(len(repn.quadratic), 1)
        self.assertEqual(repn.quadratic, {(id(m.x), id(m.y)): 1})
        assertExpressionsEqual(self, repn.nonlinear, m.x * (m.x * m.y))
        assertExpressionsEqual(
            self,
            repn.to_expression(visitor),
            m.x * m.y + m.z * m.x + 0 * m.z + m.x * (m.x * m.y),
        )



#!/usr/bin/env python
# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#
# This software is a modified version of the Kestrel interface
# package that is provided by NEOS:  http://www.neos-server.org
#

import io
import os
import re
import sys
import time
import socket
import base64
import tempfile
import logging

from pyomo.common.dependencies import attempt_import

xmlrpclib = attempt_import('xmlrpc.client')[0]
gzip = attempt_import('gzip')[0]

logger = logging.getLogger('pyomo.neos')

_email_re = re.compile(r'([^@]+@[^@]+\.[a-zA-Z0-9]+)$')


class NEOS:
    # NEOS currently only supports HTTPS access
    scheme = 'https'
    host = 'neos-server.org'
    port = '3333'
    # Legacy NEOS HTTP interface
    # urlscheme = 'http'
    # port = '3332'


def ProxiedTransport():
    from urllib.parse import urlparse
    import http.client as httplib

    # ProxiedTransport from Python 3.x documentation
    # (https://docs.python.org/3/library/xmlrpc.client.html)
    class ProxiedTransport_PY3(xmlrpclib.Transport):
        def set_proxy(self, host):
            self.proxy = urlparse(host)
            if not self.proxy.hostname:
                # User omitted scheme from the proxy; assume http
                self.proxy = urlparse('http://' + host)

        def make_connection(self, host):
            scheme = urlparse(host).scheme
            if not scheme:
                scheme = NEOS.scheme

            # Empirically, the connection class in Python 3.x needs to
            # match the final endpoint connection scheme, NOT the proxy
            # scheme.  The set_tunnel host then should NOT have a scheme
            # attached to it.
            if scheme == 'https':
                connClass = httplib.HTTPSConnection
            else:
                connClass = httplib.HTTPConnection

            connection = connClass(self.proxy.hostname, self.proxy.port)
            connection.set_tunnel(host)
            return connection

    return ProxiedTransport_PY3()


class kestrelAMPL:
    def __init__(self):
        self.setup_connection()

    def __del__(self):
        # This is a hack to force the socket to close.  When running
        # tests under unittest, unittest will report any underlying
        # socket resources that have not been closed as warnings.  This
        # will cause the socket to be closed when the kestrelAMPL object
        # falls out of scope and will suppress the warnings from
        # unittest.
        #
        # Note that this is only to suppress warnings, as __del__ is not
        # guaranteed to be called (especially for any objects that still
        # exist when the Python process terminates)
        if getattr(self, 'neos', None) is not None:
            try:
                if hasattr(self, 'transport') and hasattr(self.transport, 'close'):
                    self.transport.close()
            except Exception:
                # never raise from a destructor
                pass

    def setup_connection(self):
        import http.client

        # on *NIX, the proxy can show up either upper or lowercase.
        # Prefer lower case, and prefer HTTPS over HTTP if the
        # NEOS.scheme is https.
        proxy = os.environ.get('http_proxy', os.environ.get('HTTP_PROXY', ''))
        if NEOS.scheme == 'https':
            proxy = os.environ.get('https_proxy', os.environ.get('HTTPS_PROXY', proxy))
        if proxy:
            self.transport = ProxiedTransport()
            self.transport.set_proxy(proxy)
        elif NEOS.scheme == 'https':
            self.transport = xmlrpclib.SafeTransport()
        else:
            self.transport = xmlrpclib.Transport()

        self.neos = xmlrpclib.ServerProxy(
            "%s://%s:%s" % (NEOS.scheme, NEOS.host, NEOS.port), transport=self.transport
        )

        logger.info("Connecting to the NEOS server ... ")
        try:
            result = self.neos.ping()
            logger.info("OK.")
        except (
            socket.error,
            xmlrpclib.ProtocolError,
            http.client.BadStatusLine,
            NotImplementedError,
            Exception,
        ) as e:
            self.connect_error = e
            self.neos = None
            logger.info("Fail: %s" % (e,))
            logger.warning("NEOS is temporarily unavailable:\n\t(%s)" % (e,))

    def tempfile(self):
        ampl_id = os.getenv('ampl_id', 'unknown')
        return os.path.join(tempfile.gettempdir(), f'at{ampl_id}.jobs')

    def kill(self, jobNumber, password):
        response = self.neos.killJob(jobNumber, password)
        logger.info(response)

    def solvers(self):
        if self.neos is None:
            return []
        else:
            attempt = 0
            while attempt < 3:
                try:
                    return self.neos.listSolversInCategory("kestrel")
                except (socket.timeout, Exception):
                    attempt += 1
            return []

    def retrieve(self, stub, jobNumber, password):
        # NEOS should return results as uu-encoded xmlrpclib.Binary data
        results = self.neos.getFinalResults(jobNumber, password)
        if isinstance(results, xmlrpclib.Binary):
            results = results.data
        if isinstance(results, str):
            results = results.encode()
        # decode results to kestrel.sol; well try to anyway, any errors
        # will result in error strings in .sol file instead of solution.
        if stub[-4:] == '.sol':
            stub = stub[:-4]
        with open(stub + ".sol", "wb") as solfile:
            solfile.write(results)

    def submit(self, xml):
        # LOGNAME and USER should map to the effective user (i.e., the
        # one sudo-ed to), whereas USERNAME is the original user who ran
        # sudo.  We include USERNAME to cover Windows, where LOGNAME and
        # USER may not be defined.
        user = self.getEmailAddress()
        jobNumber, password = self.neos.submitJob(xml, user, "kestrel")
        if jobNumber == 0:
            raise RuntimeError("%s\n\tJob not submitted" % (password,))

        logger.info("Job %d submitted to NEOS, password='%s'\n" % (jobNumber, password))
        logger.info("Check the following URL for progress report :\n")
        logger.info(
            "%s://www.neos-server.org/neos/cgi-bin/nph-neos-solver.cgi"
            "?admin=results&jobnumber=%d&pass=%s\n" % (NEOS.scheme, jobNumber, password)
        )
        return (jobNumber, password)

    def getEmailAddress(self):
        # Note: the NEOS email address parser is more restrictive than
        # the email.utils parser
        email = os.environ.get('NEOS_EMAIL', '')
        if _email_re.match(email):
            return email

        raise RuntimeError(
            "NEOS requires a valid email address. "
            "Please set the 'NEOS_EMAIL' environment variable."
        )

    def getJobAndPassword(self):
        """
        If kestrel_options is set to job/password, then return
        the job and password values
        """
        jobNumber = 0
        password = ""
        options = os.getenv("kestrel_options")
        if options is not None:
            m = re.search(r'job\s*=\s*(\d+)', options, re.IGNORECASE)
            if m:
                jobNumber = int(m.groups()[0])
            m = re.search(r'password\s*=\s*(\S+)', options, re.IGNORECASE)
            if m:
                password = m.groups()[0]
        return (jobNumber, password)

    def getAvailableSolvers(self):
        """Return a list of all NEOS solvers that this interface supports"""
        if self.neos is None:
            return []
        try:
            allKestrelSolvers = self.neos.listSolversInCategory("kestrel")
        except Exception as e:
            logger.warning("Failed to retrieve solver list from NEOS: %s", e)
            return []
        _ampl = ':AMPL'
        return sorted(s[: -len(_ampl)] for s in allKestrelSolvers if s.endswith(_ampl))

    def getSolverName(self):
        """
        Read in the kestrel_options to pick out the solver name.

        The tricky parts:

          - we don't want to be case sensitive, but NEOS is.
          - we need to read in options variable
        """
        # Get a list of available kestrel solvers from NEOS
        kestrelAmplSolvers = self.getAvailableSolvers()

        NEOS_solver_name = None
        m = None

        # Read kestrel_options to get solver name
        if "kestrel_options" in os.environ:
            self.options = os.getenv("kestrel_options")
        elif "KESTREL_OPTIONS" in os.environ:
            self.options = os.getenv("KESTREL_OPTIONS")
        else:
            self.options = None

        if self.options is not None:
            m = re.search(r'solver\s*=*\s*(\S+)', self.options, re.IGNORECASE)
            if m:
                solver_name = m.groups()[0]
                for s in kestrelAmplSolvers:
                    if s.upper() == solver_name.upper():
                        NEOS_solver_name = s
                        break
                if not NEOS_solver_name:
                    raise RuntimeError(
                        "%s is not available on NEOS.  Choose from:\n\t%s"
                        % (solver_name, "\n\t".join(kestrelAmplSolvers))
                    )

        if NEOS_solver_name is None:
            raise RuntimeError(
                "%s is not available on NEOS.  Choose from:\n\t%s"
                % (solver_name, "\n\t".join(kestrelAmplSolvers))
            )
        return NEOS_solver_name

    def formXML(self, stub):
        solver = self.getSolverName()
        zipped_nl_file = io.BytesIO()
        if os.path.exists(stub) and stub[-3:] == '.nl':
            stub = stub[:-3]
        nlfile = open(stub + ".nl", "rb")
        zipper = gzip.GzipFile(mode='wb', fileobj=zipped_nl_file)
        zipper.write(nlfile.read())
        zipper.close()
        nlfile.close()
        #
        ampl_files = {}
        for key in ['adj', 'col', 'env', 'fix', 'spc', 'row', 'slc', 'unv']:
            if os.access(stub + "." + key, os.R_OK):
                with open(stub + "." + key, "r") as f:
                    ampl_files[key] = f.read()

        # Get priority
        priority = ""
        if self.options:
            m = re.search(r'priority[\s=]+(\S+)', self.options)
            if m:
                priority = "<priority>%s</priority>\n" % (m.groups()[0])

        # Add any AMPL-created environment variables to dictionary
        solver_options = "kestrel_options:solver=%s\n" % solver.lower()
        solver_options_key = "%s_options" % solver
        #
        solver_options_value = ""
        if solver_options_key in os.environ:
            solver_options_value = os.getenv(solver_options_key)
        elif solver_options_key.lower() in os.environ:
            solver_options_value = os.getenv(solver_options_key.lower())
        elif solver_options_key.upper() in os.environ:
            solver_options_value = os.getenv(solver_options_key.upper())
        if solver_options_value:
            solver_options += "%s_options:%s\n" % (solver.lower(), solver_options_value)
        #
        nl_string = (base64.encodebytes(zipped_nl_file.getvalue())).decode('utf-8')
        xml = """
              <document>
              <category>kestrel</category>
              <email>%s</email>
              <solver>%s</solver>
              <inputType>AMPL</inputType>
              %s
              <solver_options>%s</solver_options>
              <nlfile><base64>%s</base64></nlfile>\n""" % (
            self.getEmailAddress(),
            solver,
            priority,
            solver_options,
            nl_string,
        )
        #
        for key in ampl_files:
            xml += "<%s><![CDATA[%s]]></%s>\n" % (key, ampl_files[key], key)
        #
        for option in ["kestrel_auxfiles", "mip_priorities", "objective_precision"]:
            if option in os.environ:
                xml += "<%s><![CDATA[%s]]></%s>\n" % (option, os.getenv(option), option)
        #
        xml += "</document>"
        return xml


if __name__ == "__main__":  # pragma:nocover
    if len(sys.argv) < 2:
        sys.stdout.write("kestrel should be called from inside AMPL.\n")
        sys.exit(1)

    kestrel = kestrelAMPL()

    if sys.argv[1] == "solvers":
        for s in sorted(kestrel.neos.listSolversInCategory("kestrel")):
            print(" " + s)
        sys.exit(0)

    elif sys.argv[1] == "submit":
        xml = kestrel.formXML("kestproblem")
        jobNumber, password = kestrel.submit(xml)

        # Add the job,pass to the stack
        jobfile = open(kestrel.tempfile(), 'a')
        jobfile.write("%d %s\n" % (jobNumber, password))
        jobfile.close()

    elif sys.argv[1] == "retrieve":
        # Pop job,pass from the stack
        try:
            jobfile = open(kestrel.tempfile(), 'r')
        except IOError:
            e = sys.exc_info()[1]
            sys.stdout.write("Error, could not open file %s.\n")
            sys.stdout.write("Did you use kestrelsub?\n")
            sys.exit(1)

        m = re.match(r'(\d+) ([a-zA-Z]+)', jobfile.readline())
        if m:
            jobNumber = int(m.groups()[0])
            password = m.groups()[1]
        restofstack = jobfile.read()
        jobfile.close()

        kestrel.retrieve('kestresult', jobNumber, password)

        if restofstack:
            sys.stdout.write("restofstack: %s\n" % restofstack)
            jobfile = open(kestrel.tempfile(), 'w')
            jobfile.write(restofstack)
            jobfile.close()
        else:
            os.unlink(kestrel.tempfile())

    elif sys.argv[1] == "kill":
        jobNumber, password = kestrel.getJobAndPassword()
        if jobNumber:
            kestrel.kill(jobNumber, password)
        else:
            sys.stdout.write(
                "To kill a NEOS job, first set kestrel_options variable:\n"
            )
            sys.stdout.write(
                '\tampl: option kestrel_options "job=#### password=xxxx";\n'
            )
    else:
        try:
            stub = sys.argv[1]
            # See if kestrel_options has job=.. password=..
            jobNumber, password = kestrel.getJobAndPassword()

            # otherwise, submit current problem to NEOS
            if not jobNumber:
                xml = kestrel.formXML(stub)
                jobNumber, password = kestrel.submit(xml)

        except KeyboardInterrupt:
            e = sys.exc_info()[1]
            sys.stdout.write("Keyboard Interrupt while submitting problem.\n")
            sys.exit(1)
        try:
            # Get intermediate results
            time.sleep(1)
            status = "Running"
            offset = 0
            while status == "Running" or status == "Waiting":
                output, offset = kestrel.neos.getIntermediateResults(
                    jobNumber, password, offset
                )

                if isinstance(output, xmlrpclib.Binary):
                    output = output.data
                sys.stdout.write(output)
                status = kestrel.neos.getJobStatus(jobNumber, password)
                time.sleep(5)

            # Get final results
            kestrel.retrieve(stub, jobNumber, password)
            sys.exit(0)
        except KeyboardInterrupt:
            e = sys.exc_info()[1]
            msg = '''
Keyboard Interrupt\n\
Job is still running on remote machine\n\
To stop job:\n\
\tampl: option kestrel_options "job=%d password=%s";\n\
\tampl: commands kestrelkill;\n\
To retrieve results:\n\
\tampl: option kestrel_options "job=%d password=%s";\n\
\tampl: solve;\n''' % (jobNumber, password, jobNumber, password)
            sys.stdout.write(msg)
            sys.exit(1)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for util/misc
#

import os

import pyomo.common.unittest as unittest

from pyomo.opt import (
    AbstractProblemWriter,
    AbstractResultsReader,
    OptSolver,
    ReaderFactory,
    SolverFactory,
    WriterFactory,
)
from pyomo.opt.base.solvers import UnknownSolver
from pyomo.opt.plugins.sol import ResultsReader_sol
from pyomo.solvers.plugins.solvers.CBCplugin import MockCBC


class MockWriter(AbstractProblemWriter):
    def __init__(self, name=None):
        AbstractProblemWriter.__init__(self, name)


class MockReader(AbstractResultsReader):
    def __init__(self, name=None):
        AbstractResultsReader.__init__(self, name)


class MockSolver(OptSolver):
    def __init__(self, **kwds):
        kwds['type'] = 'stest_type'
        kwds['doc'] = 'MockSolver Documentation'
        OptSolver.__init__(self, **kwds)


class OptFactoryDebug(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        import pyomo.environ

    def tearDown(self):
        ReaderFactory.unregister('rtest3')
        SolverFactory.unregister('stest3')
        WriterFactory.unregister('wtest3')

    def test_solver_factory(self):
        """
        Testing the pyomo.opt solver factory with MIP solvers
        """
        SolverFactory.register('stest3')(MockSolver)
        ans = sorted(SolverFactory)
        tmp = [
            '_mock_asl',
            '_mock_cbc',
            '_mock_cplex',
            '_mock_glpk',
            'cbc',
            'cplex',
            'glpk',
            'scip',
            'stest3',
            'asl',
        ]
        tmp.sort()
        self.assertTrue(
            set(tmp) <= set(ans), msg="Set %s is not a subset of set %s" % (tmp, ans)
        )

    def test_solver_instance(self):
        """
        Testing that we get a specific solver instance
        """
        ans = SolverFactory("none")
        self.assertTrue(isinstance(ans, UnknownSolver))
        ans = SolverFactory("_mock_cbc")
        self.assertEqual(type(ans), MockCBC)
        ans = SolverFactory("_mock_cbc", name="mymock")
        self.assertEqual(type(ans), MockCBC)
        self.assertEqual(ans.name, "mymock")

    def test_solver_registration(self):
        """
        Testing methods in the solverwriter factory registration process
        """
        SolverFactory.unregister('stest3')
        self.assertTrue('stest3' not in SolverFactory)
        SolverFactory.register('stest3')(MockSolver)
        self.assertTrue('stest3' in SolverFactory)
        self.assertTrue('_mock_cbc' in SolverFactory)

    def test_writer_factory(self):
        """
        Testing the pyomo.opt writer factory with MIP writers
        """
        WriterFactory.register('wtest3')(MockWriter)
        factory = WriterFactory
        self.assertTrue(set(['wtest3']) <= set(factory))

    def test_writer_instance(self):
        """
        Testing that we get a specific writer instance

        Note: this simply provides code coverage right now, but
        later it should be adapted to generate a specific writer.
        """
        ans = WriterFactory("none")
        self.assertEqual(ans, None)
        ans = WriterFactory("wtest3")
        self.assertEqual(ans, None)
        WriterFactory.register('wtest3')(MockWriter)
        ans = WriterFactory("wtest3")
        self.assertNotEqual(ans, None)

    def test_writer_registration(self):
        """
        Testing methods in the writer factory registration process
        """
        WriterFactory.unregister('wtest3')
        self.assertTrue(not 'wtest3' in WriterFactory)
        WriterFactory.register('wtest3')(MockWriter)
        self.assertTrue('wtest3' in WriterFactory)

    def test_reader_factory(self):
        """
        Testing the pyomo.opt reader factory
        """
        ReaderFactory.register('rtest3')(MockReader)
        ans = ReaderFactory
        # self.assertEqual(len(ans),4)
        self.assertTrue(set(ans) >= set(["rtest3", "sol", "yaml", "json"]))

    def test_reader_instance(self):
        """
        Testing that we get a specific reader instance
        """
        ans = ReaderFactory("none")
        self.assertEqual(ans, None)
        ans = ReaderFactory("sol")
        self.assertEqual(type(ans), ResultsReader_sol)
        # ans = pyomo.opt.ReaderFactory("osrl", "myreader")
        # self.assertEqual(type(ans), pyomo.opt.reader.OS.ResultsReader_osrl)
        # self.assertEqual(ans.name, "myreader")

    def test_reader_registration(self):
        """
        Testing methods in the reader factory registration process
        """
        ReaderFactory.unregister('rtest3')
        self.assertTrue(not 'rtest3' in ReaderFactory)
        ReaderFactory.register('rtest3')(MockReader)
        self.assertTrue('rtest3' in ReaderFactory)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pyomo.environ as pyo
from pyomo.core.base import ConcreteModel, Var, Constraint, Objective
from pyomo.common.dependencies import attempt_import

np, numpy_available = attempt_import(
    'numpy', 'Interior point requires numpy', minimum_version='1.13.0'
)
scipy, scipy_available = attempt_import('scipy', 'Interior point requires scipy')
mumps, mumps_available = attempt_import('mumps', 'Interior point requires mumps')
if not (numpy_available and scipy_available):
    raise unittest.SkipTest('Interior point tests require numpy and scipy')
if scipy_available:
    from pyomo.contrib.interior_point.linalg.scipy_interface import ScipyInterface
if mumps_available:
    from pyomo.contrib.interior_point.linalg.mumps_interface import MumpsInterface

from pyomo.contrib.pynumero.asl import AmplInterface

asl_available = AmplInterface.available()
if not asl_available:
    raise unittest.SkipTest('Regularization tests require ASL')

from pyomo.contrib.interior_point.interior_point import (
    InteriorPointSolver,
    InteriorPointStatus,
)
from pyomo.contrib.interior_point.interface import InteriorPointInterface

from pyomo.contrib.pynumero.linalg.ma27 import MA27Interface

ma27_available = MA27Interface.available()
if ma27_available:
    from pyomo.contrib.interior_point.linalg.ma27_interface import (
        InteriorPointMA27Interface,
    )


def make_model():
    m = ConcreteModel()
    m.x = Var([1, 2, 3], initialize=0)
    m.f = Var([1, 2, 3], initialize=0)
    m.F = Var(initialize=0)
    m.f[1].fix(1)
    m.f[2].fix(2)

    m.sum_con = Constraint(expr=(1 == m.x[1] + m.x[2] + m.x[3]))

    def bilin_rule(m, i):
        return m.F * m.x[i] == m.f[i]

    m.bilin_con = Constraint([1, 2, 3], rule=bilin_rule)

    m.obj = Objective(expr=m.F**2)

    return m


def make_model_2():
    m = ConcreteModel()
    m.x = Var(initialize=0.1, bounds=(0, 1))
    m.y = Var(initialize=0.1, bounds=(0, 1))
    m.obj = Objective(expr=-m.x**2 - m.y**2)
    m.c = Constraint(expr=m.y <= pyo.exp(-m.x))
    return m


class TestRegularization(unittest.TestCase):
    def _test_regularization(self, linear_solver):
        m = make_model()
        interface = InteriorPointInterface(m)
        ip_solver = InteriorPointSolver(linear_solver)
        ip_solver.set_interface(interface)

        interface.set_barrier_parameter(1e-1)

        # Evaluate KKT matrix before any iterations
        kkt = interface.evaluate_primal_dual_kkt_matrix()
        reg_coef = ip_solver.factorize(kkt)

        # Expected regularization coefficient:
        self.assertAlmostEqual(reg_coef, 1e-4)

        desired_n_neg_evals = (
            ip_solver.interface.n_eq_constraints()
            + ip_solver.interface.n_ineq_constraints()
        )

        # Expected inertia:
        n_pos_evals, n_neg_evals, n_null_evals = linear_solver.get_inertia()
        self.assertEqual(n_null_evals, 0)
        self.assertEqual(n_neg_evals, desired_n_neg_evals)

    @unittest.skipIf(not mumps_available, 'Mumps is not available')
    def test_mumps(self):
        solver = MumpsInterface()
        self._test_regularization(solver)

    @unittest.skipIf(not scipy_available, "Scipy is not available")
    def test_scipy(self):
        solver = ScipyInterface(compute_inertia=True)
        self._test_regularization(solver)

    @unittest.skipIf(not ma27_available, 'MA27 is not available')
    def test_ma27(self):
        solver = InteriorPointMA27Interface(icntl_options={1: 0, 2: 0})
        self._test_regularization(solver)

    def _test_regularization_2(self, linear_solver):
        m = make_model_2()
        interface = InteriorPointInterface(m)
        ip_solver = InteriorPointSolver(linear_solver)

        status = ip_solver.solve(interface)
        self.assertEqual(status, InteriorPointStatus.optimal)
        interface.load_primals_into_pyomo_model()
        self.assertAlmostEqual(m.x.value, 1)
        self.assertAlmostEqual(m.y.value, pyo.exp(-1))

    @unittest.skipIf(not mumps_available, 'Mumps is not available')
    def test_mumps_2(self):
        solver = MumpsInterface()
        self._test_regularization_2(solver)

    @unittest.skipIf(not scipy_available, "Scipy is not available")
    def test_scipy_2(self):
        solver = ScipyInterface(compute_inertia=True)
        self._test_regularization_2(solver)

    @unittest.skipIf(not ma27_available, 'MA27 is not available')
    def test_ma27_2(self):
        solver = InteriorPointMA27Interface(icntl_options={1: 0, 2: 0})
        self._test_regularization_2(solver)


if __name__ == '__main__':
    #
    unittest.main()
    # test_reg = TestRegularization()
    # test_reg.test_regularize_mumps()
    # test_reg.test_regularize_scipy()
